import cv2
import numpy as np
from app.plate import Plate
from sklearn.cluster import DBSCAN

class WellDetector:
    def __init__(self, blurred_image):
        self.blurred_image = blurred_image

    def detect_wells(self, dp=1, minDist=250, param1=45, param2=30, minRadius=155, maxRadius=180):
        circles = cv2.HoughCircles(
            self.blurred_image, 
            cv2.HOUGH_GRADIENT, 
            dp, 
            minDist,
            param1=param1,
            param2=param2, 
            minRadius=minRadius, 
            maxRadius=maxRadius    
        )
        return np.round(circles[0, :]).astype("int") if circles is not None else []

    def sort_circles(self, circles):
        circles_sorted = sorted(circles, key=lambda c: c[1])
        row_threshold = 40
        rows = []
        current_row = []
        previous_y = circles_sorted[0][1] if circles_sorted else 0

        for circle in circles_sorted:
            x, y, r = circle
            if abs(y - previous_y) > row_threshold:
                rows.append(sorted(current_row, key=lambda c: c[0]))
                current_row = [circle]
                previous_y = y
            else:
                current_row.append(circle)
                previous_y = y

        if current_row:
            rows.append(sorted(current_row, key=lambda c: c[0]))

        return [circle for row in rows for circle in row]

    def group_wells_into_plates(self, circles, plate_rows=6, plate_cols=4, eps=350):
        well_centers = [(x, y) for x, y, r in circles]
        well_centers_np = np.array(well_centers)
        expected_well_count = plate_rows * plate_cols

        # Use DBSCAN to cluster the wells based on proximity
        dbscan = DBSCAN(eps=eps, min_samples=2)
        labels = dbscan.fit_predict(well_centers_np)

        # Group wells by cluster
        clusters = {}
        for label, circle in zip(labels, circles):
            if label == -1:  # Noise, wells not part of any cluster
                continue
            if label not in clusters:
                clusters[label] = []
            clusters[label].append(circle)  # Store the original circle (x, y, r)

        # Filter clusters by expected well count
        plates = []
        plate_count = 0;
        for cluster in clusters.values():
            # if len(cluster) == expected_well_count:
            plate_count += 1
            plate = Plate(
                label=f"Plate {plate_count}",
                rows=plate_rows, 
                cols=plate_cols, 
                wells=cluster)
            plates.append(plate)
        return plates

    def draw_plates(self, image, plates):
        for plate in plates:
            plate_np = np.array([(circle[0], circle[1]) for circle in plate], dtype=np.int32)

            # Calculate the bounding rectangle
            x, y, w, h = cv2.boundingRect(plate_np)

            # Draw the bounding rectangle on the image
            cv2.rectangle(image, (x, y), (x + w, y + h), (0, 255, 0), 2)

            # Show the annotated image
            cv2.imshow('Detected Plates', image)
            cv2.waitKey(0)
            cv2.destroyAllWindows()



import toml
from pathlib import Path

class ConfigManager:
    def __init__(self, config_path=Path('./config.toml')):
        self.config_path = config_path
        if self.config_path.exists():
            self.load()
        else:
            self.config = {
            "well_detector": {
                "dp": 1,
                "min_dist": 270, 
                "param1": 45,
                "param2": 20, 
                "min_radius": 120,
                "max_radius": 145,
                "eps": 350
            },
            "well_analyzer": {
                "hsv_lower_bound": (20, 18, 0),
                "hsv_upper_bound": (179, 255, 91)
            },
            "plate": {
                "grouping": True,
                "rows": 6,
                "cols": 4,
                "well_count": 24
            }
        }

    def validate(self):
        pass

    def load(self):
        with open(self.config_path, 'r') as f:
            self.config = toml.load(f)

    def update(self, data):
        self.config.update(data)
    
    def get(self, key):
        value = self.config.get(key)
        if not value:
            raise Exception(f'{key} not found in config {self.config_path}.')
        return value

    def write(self):
        with open(self.config_path, 'w') as f:
            toml.dump(self.config, f)

    def generate(self):
        if self.config_path.exists():
            raise Exception("File already exists.")
        self.write()


# Source: https://stackoverflow.com/a/59906154
import cv2
import numpy as np


class HsvThresolder:
    def __init__(self, image_path):
        self.image = cv2.imread(image_path)

    def threshold(self, width=1200):
        # Create a window
        window_name ='image (press \'q\' quit)'
        cv2.namedWindow(window_name)
        def nothing(x):
            pass


        # Create trackbars for color change
        # Hue is from 0-179 for Opencv
        cv2.createTrackbar('HMin', window_name, 0, 179, nothing)
        cv2.createTrackbar('SMin', window_name, 0, 255, nothing)
        cv2.createTrackbar('VMin', window_name, 0, 255, nothing)
        cv2.createTrackbar('HMax', window_name, 0, 179, nothing)
        cv2.createTrackbar('SMax', window_name, 0, 255, nothing)
        cv2.createTrackbar('VMax', window_name, 0, 255, nothing)

        # Set default value for Max HSV trackbars
        cv2.setTrackbarPos('HMax', window_name, 179)
        cv2.setTrackbarPos('SMax', window_name, 255)
        cv2.setTrackbarPos('VMax', window_name, 255)

        # Initialize HSV min/max values
        hMin = sMin = vMin = hMax = sMax = vMax = 0
        phMin = psMin = pvMin = phMax = psMax = pvMax = 0

        while(1):
            # Get current positions of all trackbars
            hMin = cv2.getTrackbarPos('HMin', window_name)
            sMin = cv2.getTrackbarPos('SMin', window_name)
            vMin = cv2.getTrackbarPos('VMin', window_name)
            hMax = cv2.getTrackbarPos('HMax', window_name)
            sMax = cv2.getTrackbarPos('SMax', window_name)
            vMax = cv2.getTrackbarPos('VMax', window_name)

            # Set minimum and maximum HSV values to display
            lower = np.array([hMin, sMin, vMin])
            upper = np.array([hMax, sMax, vMax])

            # Convert to HSV format and color threshold
            hsv = cv2.cvtColor(self.image, cv2.COLOR_BGR2HSV)
            mask = cv2.inRange(hsv, lower, upper)
            result = cv2.bitwise_and(self.image, self.image, mask=mask)

            # Print if there is a change in HSV value
            if((phMin != hMin) | (psMin != sMin) | (pvMin != vMin) | (phMax != hMax) | (psMax != sMax) | (pvMax != vMax) ):
                print("(hMin = %d , sMin = %d, vMin = %d), (hMax = %d , sMax = %d, vMax = %d)" % (hMin , sMin , vMin, hMax, sMax , vMax))
                phMin = hMin
                psMin = sMin
                pvMin = vMin
                phMax = hMax
                psMax = sMax
                pvMax = vMax

            display_image = cv2.resize(result, (width, int(result.shape[0] * (width / result.shape[1]))))
            # Display result image
            cv2.imshow(window_name, display_image)
            if cv2.waitKey(10) & 0xFF == ord('q'):
                break

        cv2.destroyAllWindows()
        return (hMin, sMin, vMin), (hMax, sMax, vMax)


class Plate:
    def __init__(self, label, rows, cols, wells, row_start_letter='A'):
        self.label = label
        self.rows = rows
        self.cols = cols
        self.wells = wells
        self.row_start_letter = row_start_letter

    def well_count(self):
        return self.rows * self.cols
    
    def get_well_label(self, index):
        if self.cols == 0:
            return ''
        well_row = index // self.cols
        well_col = index % self.cols   
        label = f"{chr(ord(self.row_start_letter) + well_col)}{well_row + 1}"
        return label


from app.image_processor import ImageProcessor
from app.well_detector import WellDetector
from app.well_analyzer import WellAnalyzer
from app.visualizer import Visualizer
from app.hsv_thresholder import HsvThresolder
from app.config_manager import ConfigManager
from app.plate import Plate
from pathlib import Path
import os
import click




@click.group()
def cli():
    pass

@cli.command()
@click.option(
    "-i",
    "--image",
    "image",
    type=click.Path(path_type=Path, dir_okay=False),
    multiple=False,
    help="The image file to open.",
)
@click.option(
    "-w",
    "--width",
    "width",
    type=click.INT,
    help="The desired width to display the image.",
)
@click.option(
    "-c",
    "--config",
    "config",
    type=click.Path(path_type=Path),
    help="The config to be updated with the chosen HSV range.",
)
def threshold(image, width, config):
    """Helper to determine optimal threshold values for masking items of interest"""
    thresholder = HsvThresolder(image)
    lower_bound, upper_bound = thresholder.threshold(width)
    new_min_max_threshold = {
        "well_analyzer": {
            "hsv_lower_bound": lower_bound,
            "hsv_upper_bound": upper_bound
        }
    }
    config_manager = ConfigManager(config)
    config_manager.update(new_min_max_threshold)
    config_manager.write()
    click.echo(f"Updated thresholds in {config}")

@cli.command()
@click.option(
    "-i",
    "--image",
    "images",
    type=click.Path(path_type=Path),
    multiple=False,
    help="The image file to open.",
)
@click.option('--dp', default=1, help='Inverse ratio of the accumulator resolution to the image resolution.')
@click.option('--min_dist', default=270, help='Minimum distance between the centers of detected circles.')
@click.option('--param1', default=45, help='First method-specific parameter for the edge detection.')
@click.option('--param2', default=20, help='Second method-specific parameter for the center detection.')
@click.option('--min_radius', default=120, help='Minimum circle radius to detect.')
@click.option('--max_radius', default=145, help='Maximum circle radius to detect.')
@click.option(
    "-c",
    "--config",
    "config",
    type=click.Path(path_type=Path),
    help="The config to use. Cannot be combined with individual params for configuration.",
)
@click.option(
    "-o",
    "--output",
    "output",
    type=click.Path(path_type=Path, dir_okay=True),
    multiple=False,
    help="The name of the directory to output to.",
)
def process(images, config, output, dp, min_dist, param1, param2, min_radius, max_radius):
    """Processes images"""
    
    hsv_lower_bound = (20, 18, 0)
    hsv_upper_bound =  (179, 255, 91)
    rows = 6
    cols = 4
    well_count = 24
    eps = 350
    is_plate_grouping_enabled = True

    if config:
        try:
            config_manager = ConfigManager(config)
            config_manager.load()
            
            well_detector_config = config_manager.get('well_detector')
            dp = well_detector_config.get('dp')
            min_dist = well_detector_config.get('min_dist')
            param1 = well_detector_config.get('param1')
            param2 = well_detector_config.get('param2')
            min_radius = well_detector_config.get('min_radius')
            max_radius = well_detector_config.get('max_radius')
            eps = well_detector_config.get('eps') or eps

            well_analyzer_config = config_manager.get('well_analyzer')
            hsv_lower_bound = tuple(well_analyzer_config.get('hsv_lower_bound'))
            hsv_upper_bound = tuple(well_analyzer_config.get('hsv_upper_bound'))

            plate_config = config_manager.get('plate')
            rows = plate_config.get('rows')
            cols = plate_config.get('cols')
            well_count = plate_config.get('well_count')
            is_plate_grouping_enabled = plate_config.get('grouping')
        except Exception as e:
            click.echo(e)
            click.echo(f"Failed while processing config file {config}")
    else:
        # If no config file, use CLI values with defaults
        dp = dp or 1
        min_dist = min_dist or 270
        param1 = param1 or 45
        param2 = param2 or 20
        min_radius = min_radius or 120
        max_radius = max_radius or 145

    def process_image(image_path, output):
        click.echo(f'Processing {image_path}')
        # Step 1: Load and preprocess the image
        processor = ImageProcessor(image_path)
        blurred_image = processor.get_blurred_image()

        # Step 2: Detect and sort circles (wells)
        detector = WellDetector(blurred_image)
        wells = detector.detect_wells(dp, min_dist, param1, param2, min_radius, max_radius)
       
        sorted_wells = detector.sort_circles(wells)
        if is_plate_grouping_enabled:
            plates = detector.group_wells_into_plates(sorted_wells, plate_rows=rows, plate_cols=cols, eps=eps)
        else:
            plates = [Plate(
                label="Plate 1",
                rows=0, 
                cols=0, 
                wells=sorted_wells)]

        # Step 3: Analyze duckweed for each well
        analyzer = WellAnalyzer(processor.get_original_image(), hsv_lower_bound, hsv_upper_bound)
        visualizer = Visualizer(processor.get_original_image())

        csv_out = []
        csv_out.append("Plate,Well,Area")
        if is_plate_grouping_enabled:
            for plate in plates:
                for i, (x, y, r) in enumerate(plate.wells):
                    # Analyze the duckweed in the current well
                    contours, total_area = analyzer.analyze_plant_area(x, y, r)
            
                    # Draw the circle and contours on the image
                    visualizer.draw_contours(contours)


                    label = plate.get_well_label(i)
                    visualizer.add_text(x, y, r, f"{label}: {total_area} px")

                    # add well label and area to output
                    csv_out.append(f"{plate.label},{label},{total_area}")

                visualizer.draw_circles(plate.wells)
                visualizer.draw_plate_bounding_box(plate.wells, label=plate.label)
        else:
            for i, (x, y, r) in enumerate(sorted_wells):
                    # Analyze the duckweed in the current well
                    contours, total_area = analyzer.analyze_plant_area(x, y, r)
            
                    # Draw the circle and contours on the image
                    visualizer.draw_contours(contours)

                    visualizer.add_text(x, y, r, f"well {i}: {total_area} px")

                    # add well label and area to output
                    csv_out.append(f"1,well_{i},{total_area}")

            visualizer.draw_circles(sorted_wells)

            
        os.makedirs(output, exist_ok=True)
        out_name = os.path.splitext(os.path.basename(image_path))[0]
        out_file = str(output) + "/" + out_name + ".csv"
        with open(out_file, 'w') as f:
            for line in csv_out:
                f.write(line + '\n')

        out_image = str(output) + "/" + out_name + "_annotated.png"
        visualizer.save_image(out_image)


    if images.is_dir():
        for image_path in images.iterdir():
            process_image(image_path, output)
    else:
        process_image(images, output)


@cli.command()
@click.option(
    "-f",
    "--file",
    "file",
    type=click.Path(path_type=Path),
    help="The config file path.",
)
def config(file):
    """Generates a new config file"""
    if file:
        config_manager = ConfigManager(file)
        config_manager.generate()
    else:
        config_manager = ConfigManager()
        config_manager.generate()

# # Run the program
if __name__ == "__main__":
    cli()



import numpy as np
import cv2


class WellAnalyzer:
    def __init__(self, image, hsv_lower_bound, hsv_upper_bound):
        self.image = image
        self.hsv_lower_bound = hsv_lower_bound
        self.hsv_upper_bound = hsv_upper_bound

    def create_well_mask(self, x, y, r):
        mask = np.zeros(self.image.shape[:2], dtype=np.uint8)
        cv2.circle(mask, (x, y), r, 255, -1)
        return mask

    def analyze_plant_area(self, x, y, r):
        mask = self.create_well_mask(x, y, r)
        well_region = cv2.bitwise_and(self.image, self.image, mask=mask)
        hsv_well = cv2.cvtColor(well_region, cv2.COLOR_BGR2HSV)

        mask = cv2.inRange(hsv_well, self.hsv_lower_bound, self.hsv_upper_bound)

        # Find contours of the green areas
        contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        filtered_contours = [cnt for cnt in contours if cv2.contourArea(cnt) >= 200]
        total_area = sum(cv2.contourArea(c) for c in filtered_contours)

        return filtered_contours, total_area



import cv2
import numpy as np

class ImageProcessor:
    def __init__(self, image_path):
        self.image = cv2.imread(image_path)
        self.gray_image = self._convert_to_grayscale()
        self.equalized_image = self._equalize_brightness(self.adaptive_threshold(self.gray_image))

    def _convert_to_grayscale(self):
        return cv2.cvtColor(self.image, cv2.COLOR_BGR2GRAY)

    def _equalize_brightness(self, gray_image):
        # Apply CLAHE (Contrast Limited Adaptive Histogram Equalization)
        clahe = cv2.createCLAHE(clipLimit=20, tileGridSize=(8, 8))
        return clahe.apply(gray_image)

    def get_blurred_image(self, kernel_size=(7, 7), sigma=1.5):
        return cv2.GaussianBlur(self.equalized_image, kernel_size, sigma)

    def adaptive_threshold(self, gray_image):
        return cv2.adaptiveThreshold(gray_image, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
                                 cv2.THRESH_BINARY, 11, 2)

    def get_original_image(self):
        return self.image

    def get_resized_image(self, width):
        height = int(self.image.shape[0] * (width / self.image.shape[1]))  # Maintain aspect ratio
        return cv2.resize(self.image, (width, height))


import cv2
import numpy as np
class Visualizer:
    def __init__(self, original_image):
        self.annotated_image = original_image.copy()

    def draw_circles(self, circles):
        for x, y, r in circles:
            cv2.circle(self.annotated_image, (x, y), r, (255, 0, 255), 2)

    def draw_contours(self, contours, color=(0, 0, 255), thickness=2):
        cv2.drawContours(self.annotated_image, contours, -1, color, thickness)

    def add_text(self, x, y, r, text):
        cv2.putText(self.annotated_image, text, (x - 20, y - r - 10), cv2.FONT_HERSHEY_SIMPLEX, 
                    1.0, (255, 0, 0), 2)

    def show_image(self, window_name, width):
        resized_image = cv2.resize(self.annotated_image, (width, int(self.annotated_image.shape[0] * (width / self.annotated_image.shape[1]))))
        cv2.imshow(window_name, resized_image)
        cv2.waitKey(0)
        cv2.destroyAllWindows()

    def save_image(self, file_name):
        cv2.imwrite(file_name, self.annotated_image)
    
    def draw_plate_bounding_box(self, circles, label="Plate"):
        plate_np = np.array([(circle[0], circle[1]) for circle in circles], dtype=np.int32)

        # Calculate the bounding rectangle
        x, y, w, h = cv2.boundingRect(plate_np)

        # Draw the bounding rectangle on the image
        cv2.rectangle(self.annotated_image, (x, y), (x + w, y + h), (0, 255, 0), 2)

         # Set font and calculate the position for the label
        font = cv2.FONT_HERSHEY_SIMPLEX
        font_scale = 1
        font_thickness = 3
        text_size = cv2.getTextSize(label, font, font_scale, font_thickness)[0]

        # Calculate position: top-left corner of the bounding box
        text_x = x
        text_y = y - 10 if y - 10 > 10 else y + text_size[1] + 10

        # Draw the label on the image
        cv2.putText(self.annotated_image, label, (text_x, text_y), font, font_scale, (255, 255, 0), font_thickness)


from setuptools import setup, find_packages

setup(
    name='analyzer',
    version='0.1.0',
    py_modules=['cli'],
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'Click',
        'opencv-python',
        'numpy',
        'toml',
        'scikit-learn'
    ],
    entry_points={
        'console_scripts': [
            'analyzer = analyzer.cli:cli',
        ],
    },
)